#include <math.h>
#include "efuns.h"
#include "array.h"
#include "simul_efun.h"
#include "stralloc.h"
#include "mapping.h"
#include "list.h"
#include "main.h"
#include "opcodes.h"
#include "simulate.h"

void f_push_array()
{
  int e;
  struct vector *v;
  v=sp[0].u.vec;
  v->ref++;
  pop_stack();
  for(e=0;e<v->size;e++) push_svalue(v->item+e);
  free_vector(v);
}

extern char *get_relative_offset();


#define BT_VECTOR_MASK (BT_POINTER | BT_MAPPING | BT_ALIST_PART)

struct lvalue lvalues[LVALUES];
struct lvalue *lsp=lvalues-1;

void remove_array_from_stack(struct vector *v)
{
  struct lvalue *l;

  for (l = lvalues; l <= lsp; l++)
    if (l->parent &&
	(IS_TYPE(l->parent[0],BT_VECTOR)) &&
	l->parent->u.vec==v)
      l->sval=l->parent=NULL;
}

struct lvalue *push_indexed_lvalue(struct svalue *s,struct svalue *i,int needlval)
{
  int ind,len;

  if(!s) error("Illigal indexing.\n");

#ifdef MALLOC_DEBUG
  check_sfltable();
#endif
  switch(s->type)
  {
  case T_LVALUE:
  {
    /* note that piece of code makes it impossible to use the old
     * lvalue after the new one has been made
     */
    struct svalue *s2;
    s2=s->u.lvalue->sval;
    if(!s2)
      error("Illigal lvalue request.\n");
    *s=const0;
    /* this will not work if the lvalue given wasn't the last one...*/
    lsp--;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return push_indexed_lvalue(s2,i,needlval);
  }

  case T_POINTER:
    if(i->type!=T_NUMBER)
      error("Index isn't an integer.\n");
    ind=i->u.number;
    len=s->u.vec->size;
    if(ind<0) ind+=len;
    if(ind<0 || ind>=len)
      error("Index out of range.\n");
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n");
    lsp->type=LVALUE_VECTOR_INDEX;
    lsp->parent=s;
    lsp->sval=s->u.vec->item+ind;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return lsp;

  case T_STRING:
    if(i->type!=T_NUMBER)
      error("Index isn't an integer.\n");
    ind=i->u.number;
    len=my_strlen(s);
    if(ind<0) ind+=len;
    if(ind<0 || ind>=len) ind=len;
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n");
    lsp->type=LVALUE_STRING_INDEX;
    lsp->sval=NULL;
    lsp->parent=s;
    lsp->start=ind;
    lsp->end=ind;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return lsp;

  case T_MAPPING:
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n");
    lsp->type=LVALUE_MAPPING_INDEX;
    lsp->parent=s->u.vec->item+1;
    lsp->sval = get_map_lvalue(s->u.vec, i, needlval);
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return lsp;

  case T_LIST:
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n");
    lsp->type=LVALUE_LIST_INDEX;
    lsp->parent=NULL;
    lsp->sval = get_list_lvalue(s->u.vec, i);
    return lsp;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    break;

  default:
    error("Indexing a basic type.\n");
    return 0; /* make gcc happy */
  }
}

void assign(struct lvalue *l,struct svalue *s)
{
  if(l->sval)
  {
    if(l->parent && (IS_TYPE(l->parent[0],BT_VECTOR_MASK)))
    {
      if(IS_TYPE(*s,BT_VECTOR_MASK))
      {
	if(l->parent->u.vec->ref>1)
	  if(check_for_circularity(s->u.vec,l->parent->u.vec))
	    error("Illigal assignment, cyclic arrays not supported.\n");
      }
      l->parent->u.vec->types|=1<<s->type;
    }
    assign_svalue(l->sval,s);
    return;
  }else if(l->parent && l->parent->type==T_STRING){
    char *s1,*s2;
    int len;

    if(l->start==l->end &&
	(s->type==T_INT || (s->type==T_STRING && my_strlen(s)==1)))
    {
      int ch;

      if(s->type==T_INT)
	ch=s->u.number;
      else
	ch=s->u.string[0];

      switch(l->parent->subtype)
      {
        case STRING_MALLOC:
          /* this _must_ be my lucky day... */
          l->parent->u.string[l->start]=ch;
          return;

        case STRING_CONSTANT:
        case STRING_TEMPORARY:
          if(ch)
          {
            /* let's not make a binary string.... */
            s1=string_copy(l->parent->u.string);
            l->parent->subtype=STRING_MALLOC;
            l->parent->u.string=s1;
            s1[l->start]=ch;
            return;
          }

        case STRING_SHARED:
          /* we make a new shared string */
          len=my_strlen(l->parent);
	  if(l->start>=len || l->start<0)
	    error("Index out of range.\n");
	  s1=begin_shared_string(len);
          MEMCPY(s1,l->parent->u.string,len);
          s1[l->start]=ch;
          s1=end_shared_string(s1);
          free_svalue(l->parent);
          l->parent->type=T_STRING;
          l->parent->subtype=STRING_SHARED;
          l->parent->u.string=s1;
          return;
      }
    }

    if(s->type==T_STRING)
    {
      int len2;
      /* we make a new shared string */
      len=my_strlen(l->parent);
      len2=my_strlen(s);

      if(l->start<0) l->start=0;
      if(l->start>len) l->start=len;
      if(l->end<l->start-1) l->end=l->start-1;
      if(l->end>len) l->end=len;


      len=len-l->end+l->start-1+len2;
      s2=s1=begin_shared_string(len);
  
      if(l->start>0)
      {
        MEMCPY(s2,l->parent->u.string,l->start);
        s2+=l->start;
      }

      if(len2)
      {
        MEMCPY(s2,s->u.string,len2);
        s2+=len2;
      }

      if((len2=len+s1-s2)>0)
      {
        MEMCPY(s2,l->parent->u.string+l->end+1,len2);
        s2+=len2;
      }
 
      s2=end_shared_string(s1);
      free_svalue(l->parent);
      l->parent->type=T_STRING;
      l->parent->subtype=STRING_SHARED;
      l->parent->u.string=s2;
      return;
    }
  }
  error("Error in assignment.\n");
}

void lvalue_to_svalue_no_free(struct svalue *s,struct lvalue *l)
{
#ifdef MALLOC_DEBUG
  check_sfltable();
#endif
  if(l->sval)
  {
    assign_svalue_no_free(s,l->sval);
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return;
  }else if(l->parent){

    if(l->type==LVALUE_STRING_INDEX)
    {
      s->type=T_INT;
      s->u.number=EXTRACT_UCHAR(l->parent->u.string+l->start);
      return;
    }

    if(l->type==LVALUE_RANGE)
    {
      push_svalue(l->parent);
      push_number(l->start);
      push_number(l->end);
      f_range(3,sp-3);
      assign_svalue_no_free(s,sp);
      pop_stack();
      return;
    }
  }
  error("Error when referencing lvalue.\n");
}

void lvalue_to_svalue(struct svalue *s,struct lvalue *l)
{
  free_svalue(s);
  lvalue_to_svalue_no_free(s,l);
}

void f_index()
{
  struct svalue s;
  struct lvalue *l;
  l=push_indexed_lvalue(sp-1,sp,0);
  lvalue_to_svalue_no_free(&s,l);
  lsp=l-1;
  pop_stack();
  pop_stack();
  sp++;
  *sp=s;
}

void f_indirect()
{
#ifdef DEBUG
  if (sp->type != T_LVALUE)
    fatal("Bad type to F_INDIRECT\n");
#endif
  sp++;
  lvalue_to_svalue(sp,sp[-1].u.lvalue);
  f_swap();
  pop_stack();
}

/*
 * Get address to a valid global variable.
 */
struct svalue *find_value(int num)
{
#ifdef DEBUG
  if (num >= current_object->prog->num_variables)
  {
    fatal("Illegal variable access %d(%d) (%s). See trace above.\n",
	  num, current_object->prog->num_variables,
	  current_object->prog->name);
  }
#endif
  return current_object->variables+num;
}

void f_swap()
{
  struct svalue a;
  a=*(sp-1);
  *(sp-1)=*sp;
  *sp=a;
}

void f_swap_variables()
{
  sp++;
  lvalue_to_svalue(sp,sp[-2].u.lvalue);
  sp++;
  lvalue_to_svalue(sp,sp[-2].u.lvalue);
  assign(sp[-2].u.lvalue,sp-1);
  assign(sp[-3].u.lvalue,sp);
  pop_n_elems(4);
}



void f_cast_to_object()
{
  struct object *v;
  switch(sp->type)
  {
  case T_OBJECT:
    break;
  case T_FUNCTION:
    sp->type=T_OBJECT;
    break;
  case T_STRING:
    v=find_hashed(sp->u.string);
    if(v)
    {
      pop_stack();
      push_object(v);
    }else{
      error("Couldn't find object '%s'\n",sp->u.string);
    }
    break;

  case T_INT:
  {
    /* Here I try to use a somewhat odd 'skiplist' /Profezzorn */
    int i;
    i=sp->u.number;
    pop_stack();
    for(v=obj_list;v && v->clone_number>i;)
    {
#define DOCHECK(X) if(v->X && v->X->clone_number<v->clone_number && v->X->clone_number>=i) { v=v->X; continue; }
      DOCHECK(next_heart_beat);
      DOCHECK(next_obj_hashed);
      DOCHECK(next_inv);
      DOCHECK(contains);
      DOCHECK(super);
      v=v->next_all;
    }
    if(!v || v->clone_number!=i)
    {
      push_zero();
    }else{
      push_object(v);
    }
    break;
  }
    
  default:
    error("Casting a strange value.\n");
    break;
  }
}

void f_cast_to_function()
{
  int function;
  switch(sp->type)
  {
  case T_FUNCTION:
    break;

  default:
    error("Casting a strange value.\n");
    break;

  case T_STRING:
    if(sp[0].subtype==STRING_SHARED)
      function=find_shared_string_function(sp[0].u.string,
					   current_object->prog);
    else
      function=find_function(sp[0].u.string,current_object->prog);
      
    pop_stack();
    if(function>=0)
    {
      sp++;
      sp->type=T_FUNCTION;
      sp->u.ob=current_object;
      add_ref(current_object, "(function)");
      sp->subtype=function;
    }else{
      push_number(0);
    }
  }
}
    
void f_cast_to_float()
{
  float f;
  switch(sp->type)
  {
  case T_NUMBER:
    f=(float)sp->u.number;
    pop_stack();
    push_float(f);
    break;

  case T_FLOAT:
    break;

  case T_STRING:
    f=(float)atof(sp->u.string);
    pop_stack();
    push_float(f);
    break;

  default:
    error("Casting a strange value.\n");
  }
}

void f_cast_to_int()
{
  int i;
  switch(sp->type)
  {
  case T_FLOAT:
    i=sp->u.fnum;
    pop_stack();
    push_number((int)i);
    break;

  case T_NUMBER:
    break;

  case T_STRING:
    i=atoi(sp->u.string);
    pop_stack();
    push_number(i);
    break;

  default:
    error("Casting a strange value.\n");
  }
}

void f_cast_to_regexp()
{
  regexp *r;
  switch(sp->type)
  {
  case T_REGEXP:
    break;

  case T_STRING:
    r=regcomp(sp->u.string,0);
    if(sp->subtype==STRING_SHARED)
    {
      r->str=copy_shared_string(sp->u.string);
    }else{
      r->str=make_shared_string(sp->u.string);
    }
    pop_stack();
    sp++;
    sp->u.regexp=r;
    sp->type=T_REGEXP;
    r->ref=1;
    break;

  default:
    error("Casting a strange value.\n");
  }
}

void f_cast_to_string()
{
  char buf[30];
  switch(sp->type)
  {
  case T_FLOAT:
    sprintf(buf,"%f",sp->u.fnum);
    pop_stack();
    push_new_shared_string(buf);
    break;

  case T_NUMBER:
    sprintf(buf,"%d",sp->u.number);
    pop_stack();
    push_new_shared_string(buf);
    break;

  case T_STRING:
    break;

  default:
    error("Casting a strange value.\n");
  }
}

/*
  flags:
   *
  operators:
  %d
  %s
  %f
  %c
  %n
  %[
  %%
*/

int read_set(char *match,int cnt,char *set,int match_len)
{
  int init;
  int last=0;
  int e;

  if(cnt>=match_len)
    error("Error in sscanf format string.\n");

  if(match[cnt]=='^')
  {
    for(e=0;e<256;e++) set[e]=1;
    init=0;
    cnt++;
    if(cnt>=match_len)
      error("Error in sscanf format string.\n");
  }else{
    for(e=0;e<256;e++) set[e]=0;
    init=1;
  }
  if(match[cnt]==']')
  {
    set[last=']']=init;
    cnt++;
    if(cnt>=match_len)
      error("Error in sscanf format string.\n");
  }

  for(;match[cnt]!=']';cnt++)
  {
    if(match[cnt]=='-')
    {
      cnt++;
      if(cnt>=match_len)
	error("Error in sscanf format string.\n");
      if(match[cnt]==']')
      {
	set['-']=init;
	break;
      }
      for(e=last;e<match[cnt];e++) set[e]=init;
    }
    set[last=match[cnt]]=init;
  }
  return cnt;
}

int inter_sscanf(int num_arg)
{
  char *input;
  int input_len;
  char *match;
  int match_len;
  struct svalue sval;
  int e,cnt,matches,eye,arg;
  int no_assign;
  char set[256];
  struct svalue *argp;
  
  argp=sp-num_arg+1;

  check_argument(0,T_STRING,F_SSCANF);
  check_argument(1,T_STRING,F_SSCANF);

  input=argp[0].u.string;
  input_len=my_strlen(argp+0);

  match=argp[1].u.string;
  match_len=my_strlen(argp+1);

  arg=eye=matches=0;

  for(cnt=0;cnt<match_len;)
  {
    for(;cnt<match_len;cnt++)
    {
      if(match[cnt]=='%')
      {
        if(match[cnt+1]=='%')
        {
          cnt++;
        }else{
          break;
        }
      }
      if(eye>=input_len ||input[eye]!=match[cnt])
	return matches;
      eye++;
    }
    if(cnt>=match_len) return matches;

#ifdef DEBUG
    if(match[cnt]!='%' || match[cnt+1]=='%')
    {
      fatal("Error in sscanf.\n");
    }
#endif

    cnt++;
    if(cnt>=match_len)
      error("Error in sscanf format string.\n");

    if(match[cnt]=='*')
    {
      no_assign=1;
      cnt++;
      if(cnt>=match_len)
        error("Error in sscanf format string.\n");
    }else{
      no_assign=0;
    }

    switch(match[cnt])
    {
      case 'c':
        if(eye>=input_len) return matches;
        sval.type=T_NUMBER;
	sval.subtype=NUMBER_NUMBER;
        sval.u.number=input[eye];
        eye++;
        break;

      case 'd':
        if(eye>=input_len) return matches;

        if(isdigit(input[eye]) || input[eye]=='-')
        {
          char *t;          
          sval.type=T_NUMBER;
	  sval.subtype=NUMBER_NUMBER;
          sval.u.number=STRTOL(input+eye,&t,0);
          eye=t-input;
        }else{
          return matches;
        }
        break;

      case 's':
        if(cnt+1>=match_len)
        {
          sval.type=T_STRING;
          sval.subtype=STRING_SHARED;
          sval.u.string=make_shared_binary_string(input+eye,input_len-eye);
	  break;
        }else{
          char *end_str_start;
          char *end_str_end;
          char *s=0; /* make gcc happy */
	  char *p=0; /* make gcc happy */
          int start,contains_percent_percent;

          start=eye;
          end_str_start=match+cnt+1;
          
	  s=match+cnt+1;
	test_again:
	  if(*s=='%')
	  {
	    s++;
	    if(*s=='*') s++;
	    switch(*s)
	    {
	    case 'n':
	      s++;
	      goto test_again;
	      
	    case 's':
	      error("Illigal to have to adjecent %%s.\n");
	      return 0; /* make gcc happy */
	      
	    case 'd':
	      for(e=0;e<256;e++) set[e]=1;
	      for(e='0';e<='9';e++) set[e]=0;
	      set['-']=0;
	      goto match_set;

	    case 'f':
	      for(e=0;e<256;e++) set[e]=1;
	      for(e='0';e<='9';e++) set[e]=0;
	      set['.']=set['-']=0;
	      goto match_set;

	    case '[': /* oh dear */
	      read_set(match,s-match+1,set,match_len);
	      for(e=0;e<256;e++) set[e]=!set[e];
	      goto match_set;
	    }
	  }

	  contains_percent_percent=0;
          for(e=cnt;e<match_len;e++)
	  {
            if(match[e]=='%')
	    {
	      if(match[e+1]=='%')
	      {
		contains_percent_percent=1;
	      }else{
		break;
	      }
	    }
	  }
	   
          end_str_end=match+e;

	  if(!contains_percent_percent)
	  {
	    int end_len;
	    end_len=end_str_end-end_str_start;
	    for(;eye<input_len-end_len;eye++)
	      if(!MEMCMP(input+eye,end_str_start,end_len))
		break;
	    if(eye>=input_len-end_len)
	      return matches;
	  }else{
	    for(;eye<input_len;eye++)
	    {
	      p=input+eye;
	      for(s=end_str_start;s<end_str_end;s++,p++)
	      {
		if(*s!=*p) break;
		if(*s=='%') s++;
	      }
	      if(s==end_str_end)
		break;
	    }
	    if(eye==input_len)
	      return matches;
	  }
          sval.type=T_STRING;
          sval.subtype=STRING_SHARED;
          sval.u.string=make_shared_binary_string(input+start,eye-start);
          if(s==end_str_end)
          {
            cnt=s-match-1;
            eye=p-input;
          }         
	  break;
        }

      case '[':
        if(eye>=input_len)
        {
          return matches;
        }else{

	  cnt=read_set(match,cnt+1,set,match_len);
/* I want to be able to match the empty set as well.
   if(!set[input[eye]]) return matches;
   */

	match_set:
          for(e=eye;eye<input_len && set[EXTRACT_UCHAR(input+eye)];eye++);
          sval.type=T_STRING;
          sval.subtype=STRING_SHARED;
          sval.u.string=make_shared_binary_string(input+e,eye-e);
          break;
        }

      case 'n':
        sval.type=T_NUMBER;
	sval.subtype=NUMBER_NUMBER;
        sval.u.number=eye;
        break;
    
      case 'f':
        if(eye>=input_len) return matches;

        if(isdigit(input[eye]) || input[eye]=='-' || input[eye]=='.')
        {
	  char *t;
          sval.u.fnum=(float)STRTOD(input+eye,&t);
          sval.type=T_FLOAT;
          eye=t-input;
        }else{
          return matches;
        }
        break;        

      default:
        error("Unknown sscanf token %%%c\n",match[cnt]);
    }
    matches++;

    if(!no_assign)
    {
      arg++;
      if(num_arg-2<arg)
      {
	free_svalue(&sval);
	error("Too few arguments to sscanf.\n");
      }
      assign(argp[1+arg].u.lvalue,&sval);
    }
    free_svalue(&sval);
    cnt++;
  }
  return matches;
}


